--!parent
--!encoding=utf-8

--****************************************************************************
-- @file      LSW CORE LIB: FTP
-- @author    Valentin Schmidt
-- @version   0.2
--
-- Simple FTP client based on cURL
--****************************************************************************

property $

property pHost
property pUser
property pPass
property pPort

----------------------------------------
-- @constructor
-- @todo check/load shell xtra and pscp.exe
----------------------------------------
on new (me)

  -- libs
  $.import("curl")
  $.import("regex")

  return me
end

----------------------------------------
-- Set FTP host
-- @param {string} host
-- @param {string} user
-- @param {string} pass
-- @param {string} [pass]
----------------------------------------
on setHost (me, host, user, pass, port)
  me.pHost = host
  me.pUser = user
  me.pPass = pass
  me.pPort = port
end

----------------------------------------
-- Downloads remote file
-- remoteFile, localFile: either absolute pathes, or relative to moviepath
-- @param {string} remoteFile
-- @param {string} localFile
----------------------------------------
on download (me, remoteFile, localFile)

  -- shortcuts
  CURLOPT = $.curl.CURLOPT

  if voidP(me.pPort) then port = ""
  else port = ":"&me.pPort

  if not (remoteFile starts "/") then put "/" before remoteFile
  url = "ftp://"&me.pHost&port&remoteFile

  -- get a CURL handle (xtra instance)
  ch = $.curl.init()

  -- specify options
  ch.setOption(CURLOPT.URL, url)
  if not voidP(me.pPass) then
    ch.setOption(CURLOPT.USERNAME, me.pUser)
    ch.setOption(CURLOPT.PASSWORD, me.pPass)
  end if

 -- specify target filename
  ch.setDestinationFile(localFile)

  -- returnMode: 0=return error code (=default), 1=return data
  -- => has to be 0 (or void) to download as file!
  res = ch.exec()

  return [#status:(res=0), #output: curl_error(res)]

end

----------------------------------------
-- Uploads local file
-- remoteFile, localFile: either absolute pathes, or relative to moviepath
-- @param {string} localFile
-- @param {string} remoteFile
----------------------------------------
on upload (me, localFile, remoteFile)

  -- shortcuts
  CURLOPT = $.curl.CURLOPT

  if voidP(me.pPort) then port = ""
  else port = ":"&me.pPort

  if not (remoteFile starts "/") then put "/" before remoteFile
  url = "ftp://"&me.pHost&port&remoteFile

  -- get a CURL handle (xtra instance)
  ch = $.curl.init()

  -- specify options
  ch.setOption(CURLOPT.URL, url)
  if not voidP(me.pPass) then
    ch.setOption(CURLOPT.USERNAME, me.pUser)
    ch.setOption(CURLOPT.PASSWORD, me.pPass)
  end if
  ch.setOption(CURLOPT.UPLOAD, 1)
  --ch.setOption(CURLOPT.INFILESIZE, $.file.size(localFile)) -- optional

  -- upload source
  ch.setSourceFile(localFile)

  -- returnMode: 0=return error code (=default), 1=return data
  res = ch.exec() -- we don't want to return data

  return [#status:(res=0), #output: curl_error(res)]

end

----------------------------------------
-- Uploads local file
-- @param {string} remoteDir
-- @return {string}
----------------------------------------
on list (me, remoteDir)

  url = "ftp://"&me.pHost
  if not voidP(me.pPort) then put ":"&me.pPort after url
  if not remoteDir starts "/" then put "/" before remoteDir
  if remoteDir.char[remoteDir.length]<>"/" then put "/" after remoteDir
  put remoteDir after url

  -- get a CURL handle (xtra instance)
  ch = $.curl.init()

  -- specify options
  ch.setOption($.curl.CURLOPT.URL, url)
  ch.setOption($.curl.CURLOPT.USERNAME, me.pUser)
  ch.setOption($.curl.CURLOPT.PASSWORD, me.pPass)

  -- returnMode: 0=return error code (=default), 1=return data
  res = ch.exec(1)

  -- convert bytearray to string
  if ilk(res)=#bytearray then res = res.readRawString(res.length)

  if stringP(res) then
    return me._parseListing(res) -- list (result)
  else
    -- error occured, res contains error code
    return curl_error(res) -- string (error)
  end if

end

----------------------------------------
-- Parses "ls" listing
-- @param {string} listing
-- @return {list}
----------------------------------------
on _parseListing (me, str)
  files = []
  folders = []
  lines = $.regex.match("[^\r\n]+", str)
  repeat with l in lines
    info = me._parseListingLine(l)
    if info["name"]="." or info["name"]=".." then next repeat
    if info["type"]="file" then
      files.add(info["name"])
    else
      folders.add(info["name"])
    end if
  end repeat
  return [folders, files]
end

----------------------------------------
--
----------------------------------------
on _parseListingLine (me, str)

  info = [:]
  chunks = $.regex.split("\s+", str)

--  info["rights"] = chunks[1]
--  info["number"] = chunks[2]
--  info["user"] = chunks[3]
--  info["group"] = chunks[4]
--  info["size"] = chunks[5]
--  info["month"] = chunks[6]
--  info["day"] = chunks[7]
--  info["time"] = chunks[8]

  if chunks[1].char[1]="d" then
    info["type"]="folder"
  else
    info["type"]="file"
  end if

  -- get name
  name = chunks[9]
  cnt = count(chunks)
  repeat with i = 10 to cnt
    put SPACE&chunks[i] after name
  end repeat
  info["name"] = name
  return info
end
