--!parent
--!encoding=utf-8

--****************************************************************************
-- @file      LSW CORE LIB: SCITE
-- @author    Valentin Schmidt
-- @version   0.4
--****************************************************************************

property $

property _consoleFlag

property _stopOnError
property _quitCommand

property _sciteDirExtHwnd
property _dirHwnd

property _listenCallback
property _listenTarget

property _timer

----------------------------------------
-- @constructor
----------------------------------------
on new (me)
		
	-- win only
	if $.OS<>"win" then return FALSE
		
  -- libs
  $.import("detacher")
  $.import("msg")
  $.import("regex")
  $.import("timer")
  
  me._consoleFlag = FALSE
  
  -- stop on errors in debug mode?
  if not voidP($[#debug]) then $.debug.stopOnError = TRUE
  		
  -- find SciTE director HWND (NOT HWND of main window!)
  me._sciteDirExtHwnd = me._findSciTE()
  
  -- find HWND of current director process
  me._dirHwnd = $.msg.stage_handle()
  
  -- check occasionally if SciTE is still running, otherwise exit
  me._timer = $.timer.create(2000, #_checkScite, me)
  
  -- overwrites 'put' and adds 'err' command
	$.include(me.libdir&"io_scite.ls")
	
  return me
end

----------------------------------------
-- Starts input/output to SciTE output pane
-- @param {bool} [stopOnError=false]
-- @param {string} [quitCommand]
----------------------------------------
on consoleStart (me, stopOnError, quitCommand)
	
	me._consoleFlag = TRUE
	
	me._stopOnError = stopOnError
  me._quitCommand = quitCommand
  	
	-- setup alertHook
	_player.alertHook = me
	
	-- create an asynchronous console
  consoleCreateAsync(#_consoleCallback, me, quitCommand)
  
  -- show welcome message
  stdOut("> Welcome to Director "&_player.productVersion&" / LSW "&$.VERSION&$.LF)
  if not voidP(quitCommand) then
  	stdOut("> Enter '"&quitCommand&"' to end" &$.LF, 1)
  end if
  stdOut($.LF)
    
end

----------------------------------------
-- Stops input/output to SciTE output pane
----------------------------------------
on consoleStop (me)
	consoleClose()
	me._consoleFlag = FALSE
end

----------------------------------------
-- sends command to scite
----------------------------------------
on cmd (me, cmdStr)
	cmdStr = $.replace.replace("\\", "\\", cmdStr)
	$.msg.send_data(me._sciteDirExtHwnd, cmdStr)
end

----------------------------------------
-- print to Scite's output pane (in non console mode)
----------------------------------------
on print (me, str)
	str = string(str)
	cnt = the paramCount
	if cnt>2 then
		repeat with i = 3 to cnt
			put " "&param(i) after str
		end repeat
	end if	
	me._put(str)
end

----------------------------------------
-- start listening to WM_COPYDATA messages
----------------------------------------
on listen (me, cbHandler, cbTarget)
	
	if voidP(cbTarget) then cbTarget = _movie
	me._listenCallback = cbHandler
	me._listenTarget = cbTarget

	$.msg.msg_listen([$.msg.WM_COPYDATA],[TRUE],#_listenCallback, me)
end
  
----------------------------------------
--
----------------------------------------
on _findSciTE (me)

	SDI = $.msg.register_msg("SciTEDirectorInterface")
	hwndDesktop = $.msg.get_desktop_win()
	hwnd = $.msg.get_win(hwndDesktop, $.msg.GW_CHILD)
	repeat while (hwnd)
		res = $.msg.send_msg_timeout(hwnd, SDI, 0, 0, $.msg.SMTO_NORMAL, 1000, TRUE)
		if res=SDI then return hwnd
		hwnd = $.msg.get_win(hwnd, $.msg.GW_HWNDNEXT)
	end repeat
end

----------------------------------------
-- @callback
----------------------------------------
on alertHook (me, err, msg, type)
  case (type) of
		#script:

			-- show error in output pane
			me._err(err&": "&msg)

			-- exit with errorCode 1
			if me._stopOnError then exitCode(1)

			-- otherwise continue to run
			return -1

    otherwise:
    	return 0
    	
  end case
end

----------------------------------------
-- @callback
----------------------------------------
on _consoleCallback (me, str)
  $.detacher.detach(#_consoleCallbackDetached, me, str)
end

----------------------------------------
-- detached callback
----------------------------------------
on _consoleCallbackDetached (me, str)
	if not voidP(me._quitCommand) AND str=me._quitCommand then
		stdOut("> Goodbye!"&$.LF, 1)
		exitCode(1)
	else if str starts ">" then
		delete char 1 of str
		me._put(value(str))
	else
		do(str)
	end if
end

----------------------------------------
-- @callback
----------------------------------------
on _listenCallback (me, hwnd, msg, wParam, lParam, data)

	if ilk(data)=#bytearray then data = data.readRawString(data.length)
		
	-- for non-director mode: remove :hwnd:
	delete char 1 to (2+string(me._dirHwnd).length) of data
	
	-- replace double \ in pathes
	--data = $.string.replace("\\", "\", data)
	data = $.regex.replace("\\", "\", data)
	
	-- detach callback from undebuggable xtra callback thread
  $.detacher.detach(me._listenCallback, me._listenTarget, data)
end

----------------------------------------
-- @callback
----------------------------------------
on _checkScite (me)
	
	if ($.msg.get_win_class(me._sciteDirExtHwnd)="") then
		
		$.msg.msg_listen_stop()
		
		consoleClose()
		--_movie.halt()
		exitCode(0)
	end if
end

----------------------------------------
--
----------------------------------------
on _put (me, str)
	if me._consoleFlag then
		stdOut("> " & str&$.LF, 1)
	else
		str = $.regex.replace_multi(["\\","\r\n|\r|\n"], ["\\","\n> "], str)
		$.msg.send_data(me._sciteDirExtHwnd, "extender:print"&&"> "&str)
	end if
end

----------------------------------------
--
----------------------------------------
on _err (me, str)
	if me._consoleFlag then
		stdErr("-" & str&$.LF, 1)
	else
		str = $.regex.replace_multi(["\\","\r\n|\r|\n"], ["\\","\n- "], str)
		$.msg.send_data(me._sciteDirExtHwnd, "extender:print"&&"- "&str)
	end if
end
