--!parent
--!encoding=utf-8

--****************************************************************************
-- @file      LSW CORE LIB: SFTP
-- @author    Valentin Schmidt
-- @version   0.2
-- @requires
-- - xtra("Shell")
-- - PuTTY Secure Copy client (pscp.exe, http://www.chiark.greenend.org.uk/~sgtatham/putty/download.html)
--****************************************************************************

property $

property pPscPath
property pTimeout
property pKeyPath

property pHost
property pUser
property pPass

property pKey

property pPort
property pProtocol

--property _stdout

----------------------------------------
-- @constructor
-- @todo check/load shell xtra and pscp.exe
----------------------------------------
on new (me)

  if $.os="mac" then die("Error: No Mac Version yet!")

  -- libs
  $.import("regex")
  $.import("shell")

  psc = me.libdir & "win\pscp.exe"

  if psc contains SPACE then psc = QUOTE&psc&QUOTE
  me.pPscPath = psc

  me.pKeyPath = ""
  me.pTimeout = 30 -- 30 sec.

  return me
end

----------------------------------------
-- Set Path to directory where private key files are stored
-- @param {string} host
----------------------------------------
on setKeyPath (me, keyPath)
  if the last char of keyPath<>$.PD then put $.PD after keyPath
  me.pKeyPath = keyPath
end

----------------------------------------
-- Set SFTP host
-- @param {string} host
-- @param {string} user
-- @param {string} pass
-- @param {string} [pass]
-- @param {string} [key]
----------------------------------------
on setHost (me, host, user, pass, port, key)
  me.pHost = host
  me.pUser = user
  me.pPass = pass
  me.pPort = port
  me.pKey = key
end

----------------------------------------
-- Downloads remote file
-- remoteFile, localFile: either absolute pathes, or relative to moviepath
-- @param {string} remoteFile
-- @param {string} localFile
----------------------------------------
on download (me, remoteFile, localFile)

  if voidP(me.pProtocol) then protocol=""
  else protocol = " -"&me.pProtocol

  if voidP(me.pPort) then port = ""
  else port = " -P "&me.pPort

  if localFile contains SPACE then localFile = QUOTE&localFile&QUOTE

  if not voidP(me.pKey) then
    key = me.pKeyPath & me.pKey
    if key contains SPACE then key = QUOTE&key&QUOTE
    --cmd = me.pPscPath & protocol & port && "-batch -i" && key && me.pUser&"@"&me.pHost&":"&remoteFile && localFile && "2>%TMP%\~err.txt"
    cmd = "echo n | " & me.pPscPath & protocol & port && "-i" && key && me.pUser&"@"&me.pHost&":"&remoteFile && localFile && "2>%TMP%\~err.txt"

  else
    --cmd = me.pPscPath & protocol & port && "-batch -pw" && me.pPass && me.pUser&"@"&me.pHost&":"&remoteFile && localFile && "2>%TMP%\~err.txt"
    cmd = "echo n | " & me.pPscPath & protocol & port && "-pw" && me.pPass && me.pUser&"@"&me.pHost&":"&remoteFile && localFile && "2>%TMP%\~err.txt"
  end if

  props = [:]
  props["eol"] = RETURN
  props["timeout"] = me.pTimeout
  res = $.shell.shell_cmd (cmd, props)

  if (res contains "100%") then
    return [#status:TRUE, #output: res]
  else
    return [#status:FALSE, #output: $.shell.shell_cmd("type %TMP%\~err.txt")]
  end if

end

----------------------------------------
-- Uploads local file
-- remoteFile, localFile: either absolute pathes, or relative to moviepath
-- @param {string} localFile
-- @param {string} remoteFile
----------------------------------------
on upload (me, localFile, remoteFile)

  if voidP(me.pProtocol) then protocol=""
  else protocol = " -"&me.pProtocol

  if voidP(me.pPort) then port = ""
  else port = " -P "&me.pPort

  if localFile contains SPACE then localFile = QUOTE&localFile&QUOTE

  if not voidP(me.pKey) then
    key = me.pKeyPath & me.pKey
    if key contains SPACE then key = QUOTE&key&QUOTE
    --cmd = me.pPscPath & protocol & port && "-batch -i" && key && localFile && me.pUser&"@"&me.pHost&":"&remoteFile && "2>%TMP%\~err.txt"
    cmd = "echo n | " & me.pPscPath & protocol & port && "-i" && key && localFile && me.pUser&"@"&me.pHost&":"&remoteFile && "2>%TMP%\~err.txt"

  else
    --cmd = me.pPscPath & protocol & port && "-batch -pw" && me.pPass && localFile && me.pUser&"@"&me.pHost&":"&remoteFile && "2>%TMP%\~err.txt"
    cmd = "echo n | " & me.pPscPath & protocol & port && "-pw" && me.pPass && localFile && me.pUser&"@"&me.pHost&":"&remoteFile && "2>%TMP%\~err.txt"
  end if

  props = [:]
  props["eol"] = RETURN
  props["timeout"] = me.pTimeout
  res = $.shell.shell_cmd (cmd, props)

  if (res contains "100%") then
    return [#status:TRUE, #output: res]
  else
    return [#status:FALSE, #output: $.shell.shell_cmd("type %TMP%\~err.txt")]
  end if

end

----------------------------------------
-- Uploads local file
-- @param {string} remoteDir
-- @return {string}
----------------------------------------
on list (me, remoteDir)

  if voidP(me.pPort) then port = ""
  else port = " -P "&me.pPort

  if not voidP(me.pKey) then
    key = me.pKeyPath & me.pKey
    if key contains SPACE then key = QUOTE&key&QUOTE
    --cmd = me.pPscPath & port && "-ls -batch -i" && key && me.pUser&"@"&me.pHost&":"&remoteDir && "2>%TMP%\~err.txt"
    cmd = "echo n | "&me.pPscPath & port && "-ls -i" && key && me.pUser&"@"&me.pHost&":"&remoteDir && "2>%TMP%\~err.txt"

  else
    --cmd = me.pPscPath & port && "-ls -batch -q -pw" && me.pPass && me.pUser&"@"&me.pHost&":"&remoteDir && "2>%TMP%\~err.txt"
    cmd = "echo n | "&me.pPscPath & port && "-ls -pw" && me.pPass && me.pUser&"@"&me.pHost&":"&remoteDir && "2>%TMP%\~err.txt"
  end if

  -- sync mode (blocks SciTE)
  props = [:]
  props["eol"] = RETURN
  props["timeout"] = 0 -- no timeout
  res = $.shell.shell_cmd (cmd, props)

  if res="" then
    return $.shell.shell_cmd("type %TMP%\~err.txt") -- string (error)
  else
    return me._parseListing(res) -- list (result)
  end if

  -- async mode (@TODO)
--  me._stdout = ""
--  props = [:]
--  props["eol"] = RETURN
--  props["timeout"] = 0 -- no timeout
--  props["cb_complete"] = #slotComplete
--  props["cb_stdout"] = #slotStdout
--  props["cb_target"] = me
--
--  res = $.shell.shell_cmd_thread(cmd, props)

end

----------------------------------------
-- for async mode
-- @TODO
----------------------------------------
--on slotStdout (me, str)
--  me._stdout = me._stdout & str
--end

----------------------------------------
-- for async mode
-- @TODO
----------------------------------------
--on slotComplete (me)
--  res = me._stdout
--  if res="" then
--    return $.shell.shell_cmd("type %TMP%\~err.txt") -- string (error)
--  else
--    return me._parseListing(res) -- list (result)
--  end if
--end

----------------------------------------
-- Parses "ls" listing
-- @param {string} listing
-- @return {list}
----------------------------------------
on _parseListing (me, str)

  files = []
  folders = []
  lines = $.regex.match("[^\r\n]+", str)

  lines.deleteAt(1) -- "Listing directory /"

  repeat with l in lines
    info = me._parseListingLine(l)
    if info["name"]="." or info["name"]=".." then next repeat
    if info["type"]="file" then
      files.add(info["name"])
    else
      folders.add(info["name"])
    end if
  end repeat
  return [folders, files]
end

----------------------------------------
--
----------------------------------------
on _parseListingLine (me, str)

  info = [:]
  chunks = $.regex.split("\s+", str)

  -- uncomment if information is needed:
  --  info["rights"] = chunks[1]
  --  info["number"] = chunks[2]
  --  info["user"] = chunks[3]
  --  info["group"] = chunks[4]
  --  info["size"] = chunks[5]
  --  info["month"] = chunks[6]
  --  info["day"] = chunks[7]
  --  info["time"] = chunks[8]

  if chunks[1].char[1]="d" then
    info["type"]="folder"
  else
    info["type"]="file"
  end if

  -- get name
  name = chunks[9]
  cnt = count(chunks)
  repeat with i = 10 to cnt
    put SPACE&chunks[i] after name
  end repeat
  info["name"] = name
  return info
end
