--!parent
--!encoding=utf-8

--****************************************************************************
-- @file      SciTE Sidebar main class
-- @author    Valentin Schmidt
-- @version   0.1
--****************************************************************************

global $

global widget

-- main widgets
property dialog
property tabWidget

-- widgets in tabs
property explorerTree
property ftpTree
property functionList
property bookmarkList
property projectTree

-- HWND of SciTE window
property hwndScite

-- currently loaded FTP/SFTP accounts
property ftpAccounts

-- settings
property show_explorer
property show_ftp
property show_functions
property show_bookmarks
property show_projects
property functions_sort

-- currently loaded project ini file
property projectFile

-- sidebar's TMP subfolder used for editing files on FTP/SFTP servers
property tmpDir

-- map that associates temporary FTP/SFTP downloads with their origin
property tmpFileMap

----------------------------------------
-- START
----------------------------------------
on new (me)

  -- libs
  $.import("filesystem")
  $.import("ftp")
  $.import("msg")
  $.import("ini")
  $.import("qt")
  $.import("scite")
  $.import("sftp")
  $.import("shell")
  $.import("string")
  $.import("timer")

  --$.Qt.Application.setStyle(#CleanlooksStyle)

  -- window settings
  _movie.stage.title = "~Sidebar"

  -- set path to private key files for SFTP
  $.sftp.setKeyPath($.PATH&"private-keys")

  -- set drop callback
  $.Qt.setDropHandler(#slotFileDropped, me)

  tmpFileMap = [:]

  -- get TMP dir
  tmpDirShort = getEnvVar("TMP")
  tmpDir = $.shell.shell_getLongPathName(tmpDirShort)

  -- create own folder for tmp files (remove previous tmp files/subfolders)
  put $.PD & "~sidebar" after tmpDir
  cmd = "rmdir /s /q "&tmpDirShort&$.PD&"~sidebar" -- QUOTE&tmpDir&QUOTE
  $.shell.shell_cmd(cmd)
  $.filesystem.folderCreate(tmpDir)

  -- start listening to WM_COPYDATA messages
  $.scite.listen(#slotMessage, me)

  -- find SciTE window
  hwndScite = $.msg.find_win("SciTEWindow", "")
  if not hwndScite then _movie.halt()

  -- find first child (class="SciTEWindowContent")
  hwndContent = $.msg.get_win(hwndScite, 5)

  -- find first child (class="Scintilla")
  hwndScintilla = $.msg.get_win(hwndContent, 5)

  -- load settings from sidebar ini
  info = $.ini.iniToPropList($.PATH&"sidebar.ini")

  -- default settings
  show_explorer = 1
  show_ftp = 1
  show_functions = 1
  show_bookmarks = 1
  show_projects = 1
  functions_sort = 1

  if not voidP(info["Settings"]) then
    settings = info["Settings"]
    w = integer(settings["width"])
    show_explorer = integer(settings["show_explorer"])
    show_ftp = integer(settings["show_ftp"])
    show_functions = integer(settings["show_functions"])
    show_bookmarks = integer(settings["show_bookmarks"])
    show_projects = integer(settings["show_projects"])
    functions_sort = integer(settings["functions_sort"])
  end if

  -- create default rect for explorer dialog
  rectScint = $.msg.get_win_rect(hwndScintilla)
  if voidP(w) then
    w = (show_explorer+show_ftp+show_functions+show_bookmarks+show_projects) * 64
  end if

  -- find a decent position at the right side of the Scintilla pane
  CXSIZEFRAME = $.msg.get_system_metrics(32) -- SM_CXSIZEFRAME
  CYSIZEFRAME = $.msg.get_system_metrics(33) -- SM_CYSIZEFRAME
  CXVSCROLL = $.msg.get_system_metrics(2) -- SM_CXVSCROLL
  CYCAPTION = $.msg.get_system_metrics(4) -- SM_CYCAPTION

  x = rectScint.right - w - (CXVSCROLL + 2*CXSIZEFRAME + 1)
  y = rectScint.top
  h = rectScint.height - (CYCAPTION + 2*CYSIZEFRAME)
  r = rect(x, y, x+w, y+h)

  ----------------------------------------
  -- INVISIBLE WIDGET EMBEDDED INTO SciTE
  ----------------------------------------
  widget = $.Qt.Region.new(rect(0,0,0,0), hwndContent)

  ----------------------------------------
  -- DIALOG
  ----------------------------------------
  dialog = $.Qt.Dialog.new(widget)
  dialog.setGeometry(r)
  --dialog.setWindowModality(1)

  -- window with only close button in titlebar (but without "?" help button)
--  flags = bitXor(dialog.windowFlags(), $.Qt.WindowContextHelpButtonHint)
--  dialog.setWindowFlags(flags)

  -- window without any buttons in titlebar
  flags = $.Qt.Window
  flags = bitOr(flags, $.Qt.WindowTitleHint)
  flags = bitOr(flags, $.Qt.CustomizeWindowHint)
  dialog.setWindowFlags(flags)

  dialog.setWindowTitle("Sidebar")
  dialog.setWindowIcon($.PATH&"icons/scite.png")
  dialog.show()

  -- callbacks
  dialog.connect($.Qt.Event.Close, me, #slotClosed)
  dialog.connect($.Qt.Event.KeyPress, me, #slotKeyPress)

  ----------------------------------------
  -- TABWIDGET
  ----------------------------------------

  tabWidget = $.qt.TabWidget.new(dialog)
  tabWidget.setMovable(1)
  tabWidget.setStyleSheet("QTabBar::tab{padding:3px 8px;} QTabWidget::pane{;}")

  -- callbacks
  tabWidget.connect(#currentChanged, me, #slotTabChanged)

  ----------------------------------------
  -- LAYOUT
  ----------------------------------------
  hbox = $.Qt.HBoxLayout.new()
  dialog.setLayout(hbox)
  hbox.setContentsMargins (0,0,0,0)
  hbox.addWidget(tabWidget)

  tabNum = 0

  ----------------------------------------
  -- FILE EXPLORER TREEWIDGET
  ----------------------------------------
  if show_explorer then
    explorerTree = $.Qt.TreeView.new(tabWidget, "My Computer")

    -- hide header
    explorerTree.setHeaderHidden(TRUE)

    -- hide all colums other than "name"
    explorerTree.setColumnHidden(1, TRUE)
    explorerTree.setColumnHidden(2, TRUE)
    explorerTree.setColumnHidden(3, TRUE)

    -- callbacks
    explorerTree.connect(#doubleClicked, me, #slotExplorerDoubleClicked)
    explorerTree.connect($.Qt.Event.KeyPress, me, #slotKeyPress)

    tabWidget.addTab(explorerTree, "Explorer")
    tabWidget.setTabToolTip(tabNum, "File Explorer")
    tabNum = tabNum+1
  end if

  ----------------------------------------
  -- FTP TREEWIDGET
  ----------------------------------------
  if show_ftp then
    ftpTree = $.Qt.TreeWidget.new(tabWidget)
    ftpTree.setHeaderHidden(TRUE)

    -- load INI file
    me.loadFtpIni($.PATH&"ftp.ini")

    -- callbacks
    ftpTree.connect(#itemDoubleClicked, me, #slotFtpTreeDoubleClicked)
    ftpTree.connect($.Qt.Event.KeyPress, me, #slotKeyPress)

    tabWidget.addTab(ftpTree, "FTP/SFTP")
    tabWidget.setTabToolTip(tabNum, "Edit file on FTP/SFTP server")
    tabNum = tabNum+1
  end if

  ----------------------------------------
  -- FUNCTIONS LISTWIDGET
  ----------------------------------------
  if show_functions then
    functionList = $.Qt.ListWidget.new(tabWidget)

    -- callbacks
    functionList.connect(#doubleClicked, me, #slotFunctionDoubleClicked)
    functionList.connect($.Qt.Event.KeyPress, me, #slotKeyPress)

    tabWidget.addTab(functionList, "Functions")
    tabWidget.setTabToolTip(tabNum, "Function List")
    tabNum = tabNum+1
  end if

  ----------------------------------------
  -- BOOKMARKS LISTWIDGET
  ----------------------------------------
  if show_bookmarks then
    bookmarkList = $.Qt.ListWidget.new(tabWidget)

    -- load INI file
    me.loadBookmarksIni($.PATH&"bookmarks.ini")

    -- callbacks
    bookmarkList.connect(#doubleClicked, me, #slotBookmarkDoubleClicked)
    bookmarkList.connect($.Qt.Event.KeyPress, me, #slotKeyPress)

    tabWidget.addTab(bookmarkList, "Bookmarks")
    tabWidget.setTabToolTip(tabNum, "Bookmark List")
    tabNum = tabNum+1
  end if

  ----------------------------------------
  -- PROJECTS TREEWIDGET
  ----------------------------------------
  if show_projects then

    -- create a container frame (to allow dropping files from windows desktop)
    frame = $.Qt.Frame.new(tabWidget)
    frame.setAcceptDrops(TRUE)

    projectTree = $.Qt.TreeWidget.new(frame)
    projectTree.setHeaderHidden(TRUE)

    -- create layout, so treeWidget expands to full width and height
    hbox = $.Qt.HBoxLayout.new()
    frame.setLayout(hbox)
    hbox.setContentsMargins (0,0,0,0)
    hbox.addWidget(projectTree)

    if not voidP(settings["project"]) then
      p = settings["project"]
      if integer(settings["projectRelativePath"]) then
        if p.char[1]=$.PD then delete char 1 of p
        put $.PATH before p
      end if
      -- load INI file
      me.loadProject(p)
    end if

    -- callbacks
    projectTree.connect(#itemDoubleClicked, me, #slotProjectTreeDoubleClicked)
    projectTree.connect($.Qt.Event.KeyPress, me, #slotKeyPress)

    tabWidget.addTab(frame, "Projects")
    tabWidget.setTabToolTip(tabNum, "Project Browser")
    tabNum = tabNum+1
  end if

  ----------------------------------------
  -- give keyboard focus back to SciTE
  ----------------------------------------
  res = $.msg.activate_win(hwndScite)

  return me
end

----------------------------------------
-- Shows selected treeWidgetItem in "FTP/SFTP" tab
----------------------------------------
on showFtpResult (me, treeWidgetItem, res)
  if not listP(res) then return FALSE

  -- folders
  folders = res[1]
  repeat with f in folders
    childTreeItem = $.Qt.TreeWidgetItem.new()
    childTreeItem.setText(0, f)
    childTreeItem.setData(0, $.Qt.UserRole, 0) -- store in item if file (1) or folder (0)
    childTreeItem.setIcon(0, $.PATH, TRUE)

    treeWidgetItem.addChild(childTreeItem)
  end repeat

  -- files
  files = res[2]
  repeat with f in files
    childTreeItem = $.Qt.TreeWidgetItem.new()
    childTreeItem.setText(0, f)
    childTreeItem.setData(0, $.Qt.UserRole, 1) -- store in item if file (1) or folder (0)
    childTreeItem.setIcon(0, f, TRUE)

    treeWidgetItem.addChild(childTreeItem)
  end repeat

  -- expand
  treeWidgetItem.setExpanded(TRUE)
end

----------------------------------------
-- Returns path (as lingo list) for specified treeWidgetItem
----------------------------------------
on pathListFromTreeItem (me, treeWidgetItem)
  pathList = []
  repeat while true
    name = treeWidgetItem.text(0)
    p = treeWidgetItem.parent()
    pathList.addAt(1, name)
    if p._id=0 then -- account
      return pathList
    else
      treeWidgetItem = p
    end if
  end repeat
  return pathList
end

----------------------------------------
-- Loads FTP/SFTP INI file into "FTP/SFTP" tab
----------------------------------------
on loadFtpIni (me, iniFile)

  -- remove previous account nodes
  root = ftpTree.invisibleRootItem()
  cnt = root.childCount()
  if cnt then
    repeat with i = cnt-1 down to 0
      root.removeChild(root.child(i))
    end repeat
  end if

  info = $.ini.iniToPropList(iniFile)

  ftpAccounts = info
  --ftpAccounts.sort() -- ?

  childTreeItem = $.Qt.TreeWidgetItem.new()
  childTreeItem.setText(0, "[ EDIT FTP.INI ]")
  childTreeItem.setData(0, $.Qt.UserRole, -1)
  childTreeItem.setToolTip (0, iniFile)
  ftpTree.addTopLevelItem (childTreeItem)

  cnt = ftpAccounts.count
  repeat with i = 1 to cnt

    h = ftpAccounts[i]["home"]
    if voidP(h) then
      ftpAccounts[i]["home"] = "/"
    else if h.char[h.length]<>"/" then
      ftpAccounts[i]["home"] = h & "/"
    end if

    name = ftpAccounts.getPropAt(i)

    childTreeItem = $.Qt.TreeWidgetItem.new()
    childTreeItem.setData(0, $.Qt.UserRole, 0)
    childTreeItem.setText(0, name)
    childTreeItem.setToolTip (0, ftpAccounts[i]["host"])
    childTreeItem.setIcon(0, $.PATH&"icons\server.png")

    ftpTree.addTopLevelItem (childTreeItem)

  end repeat
end

----------------------------------------
-- Loads bookmarks INI file into "Bookmarks" tab
----------------------------------------
on loadBookmarksIni (me, iniFile)

  -- remove previous bookmarks
  bookmarkList.clear()

  info = $.ini.iniToPropList(iniFile)
  bookmarks = info["Bookmarks"]
  if listP(bookmarks) then

    listItem = $.Qt.ListWidgetItem.new()
    listItem.setData($.Qt.UserRole, iniFile)
    listItem.setText("[ EDIT BOOKMARKS.INI ]")
    listItem.setToolTip($.PATH&"bookmarks.ini")
    bookmarkList.addItem(listItem)

    cnt = bookmarks.count
    repeat with i = 1 to cnt
      name = bookmarks.getPropAt(i)
      path = bookmarks[i]

      listItem = $.Qt.ListWidgetItem.new()
      listItem.setData($.Qt.UserRole, path)
      listItem.setText(name)
      listItem.setToolTip(path)
      listItem.setIcon(path, TRUE)
			
      bookmarkList.addItem(listItem)

    end repeat
  end if
end

----------------------------------------
-- Loads project INI file into projects tab
----------------------------------------
on loadProject (me, iniFile)

  projectFile = iniFile
  projectPath = $.file.getFilePath(projectFile)

  -- remove previous project
  root = projectTree.invisibleRootItem()
  cnt = root.childCount()
  if cnt then
    repeat with i = cnt-1 down to 0
      root.removeChild(root.child(i))
    end repeat
  end if

  childTreeItem = $.Qt.TreeWidgetItem.new()
  childTreeItem.setText(0, "[ EDIT PROJECT ]")
  childTreeItem.setData(0, $.Qt.UserRole, projectFile)
  childTreeItem.setToolTip (0, projectFile)
  projectTree.addTopLevelItem(childTreeItem)

  -- get project name
  name = $.file.getFileName(projectFile)
  name = name.char[1..name.length-$.file.getFileType(projectFile).length-1]

  -- read project settings
  info = $.ini.iniToPropList(projectFile)
  if not voidP(info["Name"]) then name = info["Name"]

  projectTreeItem = $.Qt.TreeWidgetItem.new()
  projectTreeItem.setText(0, name)
  projectTree.addTopLevelItem(projectTreeItem)

  -- groups (folders)
  od = _player.itemDelimiter
  _player.itemDelimiter = "/"
  dirMap = [:]
  repeat with dir in info["Group"]

    cnt = dir.item.count

    dirTreeItem = $.Qt.TreeWidgetItem.new()
    dirTreeItem.setText(0, dir.item[cnt])
    dirTreeItem.setIcon(0, $.PATH, TRUE)

    if cnt>1 then
      parentTreeItem = dirMap[dir.item[1..cnt-1]]
      parentTreeItem.addChild(dirTreeItem)
    else
      projectTreeItem.addChild(dirTreeItem)
    end if

    dirMap[dir] = dirTreeItem

  end repeat
  _player.itemDelimiter  = od

  -- files in root folder
  repeat with f in info["Files"]
    name = $.file.getFileName(f)
    fileTreeItem = $.Qt.TreeWidgetItem.new()
    fileTreeItem.setData(0, $.Qt.UserRole, projectPath & $.PD & f)
    fileTreeItem.setText(0, name)
    fileTreeItem.setIcon(0, projectPath & $.PD & f, TRUE)
    projectTreeItem.addChild(fileTreeItem)
  end repeat

  -- files in subfolders (groups)
  cnt = dirMap.count
  repeat with i = 1 to cnt
    dir = dirMap.getPropAt(i)
    if listP(info["Files "&dir]) then
      repeat with f in info["Files "&dir]
        name = $.file.getFileName(f)
        fileTreeItem = $.Qt.TreeWidgetItem.new()
        fileTreeItem.setData(0, $.Qt.UserRole, projectPath & $.PD &f)
        fileTreeItem.setText(0, name)
        fileTreeItem.setIcon(0, projectPath & $.PD & f, TRUE)
        dirMap[i].addChild(fileTreeItem)
      end repeat
    end if
  end repeat
end

--**************************************
-- CALLBACKS
--**************************************

----------------------------------------
-- Messages received from SciTE (via lua backend)
-- @callback
----------------------------------------
on slotMessage (me, msg)

  if msg starts "functions:" then
    if show_functions then
      delete char 1 to 10 of msg
      handlerList = $.string.explode(",", msg)
      if functions_sort then handlerList.sort()

      -- update functionList
      functionList.clear()
      repeat with str in handlerList
        functionList.addItem(str)
      end repeat
    end if

  else if msg starts "closed:" then
    if show_ftp then
      delete char 1 to 7 of msg
      if msg starts tmpDir then
        delete char 1 to tmpDir.length+1 of msg

        -- remove from map
        tmpFileMap.deleteProp(msg)

        -- delete associated tmp file and subfolder
        $.filesystem.fileDelete(tmpDir&$.PD&msg)
        tmp = $.string.explodeBySingleChar($.PD,msg,2)
        subFolder = tmp[1]
        $.filesystem.folderDelete(tmpDir&$.PD&subFolder)
      end if
    end if

  else if msg starts "saved:" then
    delete char 1 to 6 of msg
    fn = msg

    if fn=$.PATH&"ftp.ini" then

      -- ftp.ini was edited, reload FTP tab
      me.loadFtpIni(fn)

    else if fn=$.PATH&"bookmarks.ini" then

      -- bookmarks.ini was edited, reload bookmarks tab
      me.loadBookmarksIni(fn)

    else if fn=projectFile then

      -- current project INI was edited, reload projects tab
      me.loadProject(fn)

    else if fn starts tmpDir & $.PD then

      -- FTP/SFTP tmp file was changed, upload changes to server
      delete char 1 to tmpDir.length+1 of fn
      info = tmpFileMap[fn]
      if not voidP(info) then
        acct = ftpAccounts[info.acct]
        provider = $[acct["protocol"]]
        if voidP(provider) then return FALSE

        provider.setHost(acct["host"], acct["user"], acct["pass"], acct["port"], acct["key"])

        -- upload (me, localFile, remoteFile)
        res = provider.upload(tmpDir & $.PD & fn, acct["home"]&info.path)

        if res.status=1 then
          put "File changes saved on server"
        else
          err "Error saving file on server:" && res.output
        end if
      end if
    end if

  else if msg starts "show:" then
    delete char 1 to 5 of msg

    explorerTree.setExpanded(msg, TRUE)
    explorerTree.selectPath(msg)

    -- to actually scroll to the directory, we need a timeout
    $.timer.create(1000, #slotExplorerScrollTo, me, TRUE, msg)

  else
    err "Unknown message received:"&&msg
  end if
end

----------------------------------------
-- @callback
----------------------------------------
on slotExplorerDoubleClicked (me, widgetID, tPath)

  -- open selected file in SciTE
  cmd = "open:"&tPath
  $.scite.cmd(cmd)

  -- give keyboard focus back to SciTE
  $.msg.activate_win(hwndScite)
end

----------------------------------------
-- @callback
----------------------------------------
on slotBookmarkDoubleClicked (me, widgetID, index)

  p = bookmarkList.item_(index).data($.Qt.UserRole)
	
	is_dir = (p.char[p.length]=$.PD)
	if is_dir then delete the last char of p
	
  --res = $.filesystem.fileExists(p)

  --if res=2 then -- it's a directory (and exists)
  if is_dir then -- it's a directory (and exists)

		-- check if directory exists, otherwise try to create it
    ok = $.filesystem.folderExists(p)
    if not ok then ok = $.filesystem.folderCreate(p)
    	
    if ok then
	    -- open it in explorer tab
	    tabWidget.setCurrentIndex(0)
	
	    explorerTree.setExpanded(p, TRUE)
	    explorerTree.selectPath(p)
	
	    -- to actually scroll to the directory, we need a timeout
	    $.timer.create(1000, #slotExplorerScrollTo, me, TRUE, p)
    else
      err "Directory '"&p&"' doesn't exist and couldn't be created"
    end if
    
  else -- assume it's a file

    -- check if file exists, otherwise try to create it
    ok = $.file.exists(p)
    if not ok then ok = $.file.putString(p, "")

    if ok then
      -- open selected file in SciTE
      cmd = "open:"&p
      $.scite.cmd(cmd)
    else
      err "File '"&p&"' doesn't exist and couldn't be created"
    end if
  end if

  -- give keyboard focus back to SciTE
  $.msg.activate_win(hwndScite)
end

----------------------------------------
-- @callback
----------------------------------------
on slotClosed (me, widgetID)
  exitCode(0)
end

----------------------------------------
-- @callback
----------------------------------------
on slotFunctionDoubleClicked (me, widget, index)
  functionName = functionList.item_(index).text()

  cmd = "extender:onFunctionSelected "&functionName
  $.scite.cmd(cmd)

  -- give keyboard focus back to SciTE
  $.msg.activate_win(hwndScite)
end

----------------------------------------
-- @callback
----------------------------------------
on slotTabChanged (me, widget, index)
  if show_functions then
    tabName = tabWidget.tabText(index)
    if tabName="Functions" then
      cmd = "extender:sidebarUpdateFunctions"
      $.scite.cmd(cmd)
    end if
  end if
end

----------------------------------------
-- #itemDoubleClicked -> string firstColText, string firstColUserData
-- @callback
----------------------------------------
on slotFtpTreeDoubleClicked (me, widgetID, firstColText, firstColUserData)

  treeWidgetItem = ftpTree.currentItem()

  itemData = treeWidgetItem.data(0, $.Qt.UserRole)
  if itemData = -1 then
    -- open ftp.ini in SciTE
    cmd = "open:"&$.PATH&"ftp.ini"
    $.scite.cmd(cmd)
    return
  end if

  -- get path
  pathList = me.pathListFromTreeItem(treeWidgetItem)
  acctName = pathList[1]
  pathList.deleteAt(1)

  acct = ftpAccounts[acctName]
  provider = $[acct["protocol"]] -- currently supported protocols: ftp, sftp
  if voidP(provider) then 
  	err "Unsupported protocol"
  	return FALSE
	end if
	
  is_file = (itemData=1)

  if is_file then

    -- download
    path = $.string.implode("/", pathList)

    --check if file is already opened in scite
    cnt = tmpFileMap.count
    repeat with i = 1 to cnt
      info = tmpFileMap[i]
      if info.acct=acctName AND info.path=path then
        -- re-open/activate tmp file in scite
        tmpFileName = tmpFileMap.getPropAt(i)
        cmd = "open:"&tmpDir & $.PD & tmpFileName
        $.scite.cmd(cmd)
        return
      end if
    end repeat

    -- otherwise download as new tmp file

    -- create its own subfolder with (hopefully unique, todo) random name
    tmpSubDir = "~"&random(the maxinteger)
    $.filesystem.folderCreate(tmpDir & $.PD & tmpSubDir)

    -- save as tmp file in tmpSubDir
    name = pathList[pathList.count]
    tmpFileName = tmpSubDir & $.PD & name

    res = provider.download(acct["home"]&path, tmpDir & $.PD & tmpFileName)
    if res.status then

      -- open tmp file in scite
      cmd = "open:"&tmpDir & $.PD & tmpFileName
      $.scite.cmd(cmd)

      -- add downloaded file to map
      tmpFileMap[tmpFileName] = [#acct:acctName, #path:path]

    else

      -- error occured
      err res.output

    end if

  else -- folder

    -- remove previous children (=reload)
    cnt = treeWidgetItem.childCount()
    if cnt then
      repeat with i = cnt down to 1
        c = treeWidgetItem.child(i-1)
        treeWidgetItem.removeChild(c)
      end repeat
    end if

    -- get listing
    provider.setHost(acct["host"], acct["user"], acct["pass"], acct["port"], acct["key"])
    path = $.string.implode("/", pathList)
    dir = acct["home"]&path
    
    put "Trying to connect to"&&acct["host"]&"..."
    res = provider.list(dir) -- [folders, files] or error-string
    
    if listP(res) then
    	put "OK"
    	me.showFtpResult(treeWidgetItem, res)
  	else 
  		err res
  	end if

  end if
end

----------------------------------------
-- @callback
----------------------------------------
on slotExplorerScrollTo (me, p)
  explorerTree.scrollTo(p, 1)
end

----------------------------------------
-- @callback
----------------------------------------
on slotProjectTreeDoubleClicked (me, widgetID, firstColText, firstColUserData)

  -- check if this a top level item, i.e a FTP account?
  treeWidgetItem = projectTree.currentItem()

  fn = firstColUserData
  if fn<>EMPTY then

    -- check if file exists, otherwise try to create it
    ok = $.file.exists(fn)
    if not ok then ok = $.file.putString(fn, "")

    if ok then
      -- open in SciTE
      cmd = "open:"&fn
      $.scite.cmd(cmd)
    else
      err "File '"&fn&"'doesn't exist and couldn't be created"
    end if

    -- give keyboard focus back to SciTE
    $.msg.activate_win(hwndScite)

  end if
end

----------------------------------------
-- @callback
----------------------------------------
on slotFileDropped (me, resList)
  prop = "application/x-qt-windows-mime;value="&QUOTE&"FileNameW"&QUOTE
  data = resList[prop]
  if not voidP(data) then
    data.position = 1
    fn = data.readRawString(data.length, "utf-16")
    if $.file.getFileType(fn)="ini" then
      me.loadProject(fn)
    end if
  end if
end

----------------------------------------
-- @callback
----------------------------------------
on slotKeyPress (me, widget, code, modifier)

  -- hide Sidebar when Alt+E was pressed
  if code=69 AND modifier=$.Qt.AltModifier then
    $.msg.show_win(dialog.winId(), $.msg.SW_HIDE)
  end if

  -- alternative: forward (all?) key events to SciTE?

end
