require 'winapi'

----------------------------------------
-- Continues comment lines, until empty comment
----------------------------------------
local function lingoAutoContinueComment(line)

  -- block comment continue
  if line:find("^%s*%-%-%s*$") then --line is only '--', terminate comment
    editor:LineUp()
    editor:LineDelete()
  elseif line:find("^%s*%-%-") then --line is comment, auto continue it
    editor:AddText("-- ")
  end

  return false --allow other methods to respond to event
end

----------------------------------------
-- Show Director Help CHM (if word selected, navigate to corresponding index entry)
----------------------------------------
function lingoShowHelp ()
  local parms
  if props['CurrentWord']=="" then
    parms = '-DirHelp '..props['SciteDefaultHome']..'\\help\\D11.5_help.chm'
  else
    parms = '-DirHelp -#klink '..props['CurrentWord']..' '..props['SciteDefaultHome']..'\\help\\D11.5_help.chm'
  end
  winapi.shell_exec ('open', 'KeyHH.exe', parms, props['SciteDefaultHome']..'\\tools')
end

----------------------------------------
-- Shows clickable list of handlers in current script (as autocompletion list)
----------------------------------------
function lingoHandlerList()
  local list = {}
  local code = "\r\n"..editor:GetText()
  for res in string.gmatch(code, "[\r\n]on ([%a%d_]+)") do
    table.insert(list, res)
  end
  table.sort(list)
  scite_UserListShow(list, 1, lingoOnHandlerSelected)
end

----------------------------------------
-- 
----------------------------------------
function lingoOnHandlerSelected(name)
	
	local code = editor:GetText()
	local search = "on "..name

  if string.sub(code, 1, string.len(search))==search then
    pos = 0
  else
    pos = string.find(code, "[\n\r]"..search)
  end
  if pos then

    -- goto line
    local lineNum = editor:LineFromPosition(pos)
    editor:GotoLine(lineNum)

    -- select line
    posEnd = editor.LineEndPosition[lineNum]
    editor:SetSel(pos, posEnd) -- Select a range of text.
    editor.FirstVisibleLine = lineNum - 3
  end
	
end

----------------------------------------
-- 
----------------------------------------
local function _countQuotes (str)
	local cnt = 0
	for i in string.gfind(str, '"') do
		cnt = cnt + 1
	end
	return cnt
end

----------------------------------------
--
----------------------------------------
local function _stripComment (strLine)

--	local pos = string.find (strLine, "--", 1, true)
--	if pos==nil then return strLine end
--	return trim(string.sub(strLine, 1, pos-1))
	
-- better/ but slower version: checks if "--" is not 
-- inside string, by counting QUOTES before pos

	local pos, tmp
	local startPos = 1
	while true do
		pos = string.find(strLine, "--", startPos, true)
		if pos==nil then return strLine end
		
		tmp = string.sub(strLine, 1, pos-1)
		if _countQuotes(tmp) % 2 == 0 then return trim(tmp) end
		startPos = pos + 2
	end

end

----------------------------------------
-- Re-indents/fromats current Lingo script
----------------------------------------
function lingoReindent ()
  local code = editor:GetText()
  if code=="" then return end

  -- get editor's current EOL (0=CRLF, 1=CR, 2=LF)
  local eol
  if editor.EOLMode==0 then
    eol = '\r\n'
  elseif editor.EOLMode==1 then
    eol = '\r'
  else
    eol = '\n'
  end
  local eolLen = string.len(eol)

  -- get editor's current indent type
  local tabStr
  if editor.UseTabs then
    tabStr = '\t'
  else
    tabStr = string.rep(" ", editor.Indent)
  end
  local tabLen = string.len(tabStr)

  local res = ""
  local indent = ""
  local firstWord

  for line in getLines(code, eol) do

    -- trim line
    line = trim(line)

    -- get first word
    firstWord = string.match(line, "%w+")

    if firstWord=="on" or firstWord=="case" or firstWord=="repeat" then
      line = indent..line
      indent = tabStr..indent

    elseif firstWord=="end" then
      indent = string.sub(indent, tabLen+1)
      line = indent..line

    elseif firstWord=="if" then
      line = indent..line
      local tmp = _stripComment(line)
       if string.sub(tmp, -4) == "then" then
        indent = tabStr..indent
      end
      
    elseif firstWord=="else" then
    	local tmp = _stripComment(line)
    	if tmp=="else" then
    		line = string.sub(indent, tabLen+1)..line
    	else
    		line = indent..line
    	end
      
    elseif firstWord=="tell" then
      line = indent..line
      local tmp = _stripComment(line)
      if not string.find(tmp, " to ") then
        indent = tabStr..indent
      end

    else
      line = indent..line

    end

    res = res..line..eol
  end

  if res~=code then editor:SetText(res) end
end


-- ADD MENU ITEMS --

scite_Command {
	'Director Help|lingoShowHelp|F1',
  'Handlers|lingoHandlerList|*.ls;*.lsw|F2',
  'Re-indent Script|lingoReindent|*.ls;*.lsw|Ctrl+Tab'
}

-- EVENT HANDLERS --

--scite_OnOpen(fixCode)
--scite_OnBeforeSave(stripTrailSpaces)
scite_OnEditorLine(lingoAutoContinueComment)
