--!parent
--!encoding=utf-8

--****************************************************************************
-- @file      MetadataEditor - main class
-- @author    Valentin Schmidt
-- @version   0.3
--****************************************************************************

global $

property pAppName

-- widgets
property pMainWindow
property pMenuBar
property pTabWidget

property pTableWidgets
property pTabs

-- TODO
--property pContextMenu

-- other vars
property pFileTypes
property pFileTypeMap
property pParsers
property pNamespaces

-- current media file
property pFileName

-- actions
property pActionSave
property pActionImportMetadata
property pActionExportMetadata

property pIsModified

-- inline datetimeedit widget
property pDateTimeWidget
property pDateTimeEditContainer

----------------------------------------
-- @constructor
----------------------------------------
on new(me)

  -- libs
  $.import("date")
  $.import("file")
  $.import("ini")
  $.import("qt")
  $.import("string")

  me.pAppName = "MetadataEditor"
  me.pIsModified = FALSE

  me.pFileTypes = []
  --me.pFileTypes.sort()
  me.pFileTypeMap = [:]
  me.pTableWidgets = []

  -- read prefs from INI file
  settings = $.ini.iniToPropList($.PATH&"MetadataEditor.ini")

  -- load tabs settings
  me.pTabs = [:]
  cnt = settings["tabs"].count
  repeat with i = 1 to cnt
    name = settings["tabs"].getPropAt(i)
    tabs = $.string.explodeBySingleChar(",", settings["tabs"][i])
    me.pTabs[name] = tabs
  end repeat

  -- load parsers settings
  me.pParsers = [:]
  cnt = settings["parsers"].count
  repeat with i = 1 to cnt
    name = settings["parsers"].getPropAt(i)
    me.pParsers[name] = $.include($.PATH&"code/parsers/"&name&".ls").new()
    types = $.string.explodeBySingleChar(",", settings["parsers"][i])
    repeat with t in types
      me.pFileTypes.add(t)
      me.pFileTypeMap[t] = name
    end repeat
  end repeat

  -- load namespace settings
  me.pNamespaces = [:]
  cnt = settings["ns"].count
  repeat with i = 1 to cnt
    ns = settings["ns"].getPropAt(i)
    fields = settings["ns:"&ns]
    if voidP(fields) then fields = [:]

    tmp = $.string.explodeBySingleChar(",", settings["ns"][i])
    me.pNamespaces[ns] = [#name:tmp[1], #url:tmp[2], #fields:fields]
  end repeat

  -- set Qt drop callback
  $.Qt.setDropHandler(#dropHandler, me)

  -- create main window
  me.pMainWindow = $.Qt.MainWindow.new(VOID, TRUE) -- ignore close
  me.pMainWindow.connect($.Qt.Event.Close, me, #slotQuit)
  me.pMainWindow.setAcceptDrops(TRUE)
  me.pMainWindow.setWindowTitle(me.pAppName)
  me.pMainWindow.setWindowIcon($.@("assets/appicon.png"))
  me.pMainWindow.setToolButtonStyle($.Qt.ToolButtonFollowStyle)

  -- create TabWidget
  me.pTabWidget = $.Qt.TabWidget.new(me.pMainWindow)
  me.pMainWindow.setCentralWidget(me.pTabWidget)

  -- create default tab
  labelWidget = $.Qt.Label.new(me.pTabWidget)
  labelWidget.setAlignment(bitOr($.Qt.AlignHCenter,$.Qt.AlignVCenter))
  labelWidget.setText("<b>Drop a media file into this window!</b><br />"&\
  "(Supported formats: "&$.string.implode(", ",me.pFileTypes)&")")
  me.pTabWidget.addTab(labelWidget, "Start")

  -- create menubar
  me.pMenuBar = $.Qt.MenuBar.new(me.pMainWindow)
  me.pMainWindow.setMenuBar(me.pMenuBar)

  me.setupFileActions()
  me.setupHelpActions()

  -- create statusbar
  statusBar = $.Qt.StatusBar.new(pMainWindow)
  me.pMainWindow.setStatusBar(statusBar)

  -- create context-menu
  --me.setupContextMenu()

  -- load and apply saved window state
  settings = $.Qt.Settings.new("Fluxus", me.pAppName, $.Qt.QSettings.IniFormat)
  state = value(settings.value("mainwindow/state"))
  if state=2 then
    me.pMainWindow.resize(320, 480)
    me.pMainWindow.setWindowState(state)
  else
    rect = value(settings.value("mainwindow/rect"))
    if ilk(rect)=#rect then me.pMainWindow.setGeometry(rect)
    else me.pMainWindow.resize(680, 680)
  end if

  -- check for command line arguments (files dropped to app icon)
  if count($.ARGS) then
    fn = $.ARGS[1]
    if $.file.getFileType(fn)<>"lsw" then
      if $.OS="mac" then fn = qx_posix2hfs(fn)
      me.load(fn)
    end if
  end if

  -- show main window
  me.pMainWindow.show()

  return me
end

--------------------------------------
--
--------------------------------------
on createTable (me, ns)

 -- create TableWidget
  tableWidget = $.Qt.TableWidget.new(me.pTabWidget)

  rowCnt = ns.fields.count

  -- configure tablewidget
  tableWidget.setColumnCount(2)
  tableWidget.setRowCount(rowCnt)
  tableWidget.setHorizontalHeaderLabels(["Property", "Value"])
  tableWidget.setResizeMode($.Qt.QHeaderView.Stretch)
  tableWidget.setVerticalHeaderVisible(FALSE)
  tableWidget.setHorizontalHeaderDefaultAlignment(1)

  -- optional
  --tableWidget.setSortingEnabled(TRUE)
  --tableWidget.setAlternatingRowColors(TRUE)

  readOnly = ns.name contains "read-only"

  -- add table rows
  repeat with i = 1 to rowCnt
    --tItem = tableWidget.item_(0, i-1)

    -- create property item (read-only)
    tItem = $.Qt.TableWidgetItem.new(tableWidget)

    -- flags = bitOr(tItem.flags(), bitNot($.Qt.ItemIsEditable))
    flags = bitNot($.Qt.ItemIsEditable)
    tItem.setFlags(flags)
    tItem.setText(ns.fields.getPropAt(i))
    tableWidget.setItem (i-1,0,tItem)

    -- create value item
    tItem = $.Qt.TableWidgetItem.new(tableWidget)

    -- make date fields read-only
    if readOnly OR ns.fields[i]="date" then
      tItem.setFlags(bitNot($.Qt.ItemIsEditable))
    end if

    tableWidget.setItem (i-1,1,tItem)

  end repeat

  -- setup TableWidget signals
  tableWidget.connect(#cellDoubleClicked, me, #cellDoubleClicked)

  -- setup context menu
  --tableWidget.setContextMenuPolicy($.Qt.CustomContextMenu)
  --tableWidget.connect(#customContextMenuRequested, me, #customContextMenuRequested)

  return tableWidget
end

--------------------------------------
--
--------------------------------------
--on setupContextMenu me
--  me.pContextMenu = $.Qt.Menu.new(me.pMainWindow)
--
-- -- action "Add Child"
--  action = $.Qt.Action.new(me.pContextMenu)
--  action.setText("Add new Tag")
--  action.connect(#triggered, me, #slotAddTag)
--  me.pContextMenu.addAction(action)
--
--  -- action "Delete"
--  action = $.Qt.Action.new(me.pContextMenu)
--  action.setText("Delete Tag")
--  action.connect(#triggered, me, #slotDeleteTag)
--  me.pContextMenu.addAction(action)
--end

--------------------------------------
--
--------------------------------------
on setupFileActions(me)

  --toolBar = $.Qt.ToolBar.new(me.pMainWindow)
  --toolBar.setWindowTitle("File Actions")
  --me.pMainWindow.addToolBar(toolBar)

  menu = me.pMenuBar.addMenu("&File")

  -- OPEN
  action = $.Qt.Action.new(menu)
  action.setText("&Open Media File...")
  --action.setIcon(me.pIconPath & "filenew.png")
  action.connect(#triggered, me, #slotFileOpen)
  action.setShortcut("Ctrl+O")
  menu.addAction(action)
  --  toolBar.addAction(action)

  -- NEW
  action = $.Qt.Action.new(menu)
  action.setText("&Save Changes")
  --action.setIcon(me.pIconPath & "filenew.png")
  action.connect(#triggered, me, #slotFileSave)
  action.setShortcut("Ctrl+S")
  action.setEnabled(FALSE)
  menu.addAction(action)
  --  toolBar.addAction(action)
  me.pActionSave = action

  menu.addSeparator()

  action = $.Qt.Action.new(menu)
  action.setText("&Import Metadata...")
  --action.setIcon(me.pIconPath & "filenew.png")
  action.connect(#triggered, me, #slotFileImportMetadata)
  action.setEnabled(FALSE)
  menu.addAction(action)
  me.pActionImportMetaData = action

  action = $.Qt.Action.new(menu)
  action.setText("&Export Metadata...")
  --action.setIcon(me.pIconPath & "filenew.png")
  action.connect(#triggered, me, #slotFileExportMetadata)
  action.setEnabled(FALSE)
  menu.addAction(action)
  me.pActionExportMetaData = action

  menu.addSeparator()

  action = $.Qt.Action.new(menu)
  action.setText("&Load CSV (Batch-Processing)...")
  --action.setIcon(me.pIconPath & "filenew.png")
  action.connect(#triggered, me, #slotFileLoadCSV)
  --action.setEnabled(FALSE)
  menu.addAction(action)

  menu.addSeparator()

  -- QUIT
  action = $.Qt.Action.new(menu)
  action.setText("&Quit")
  action.connect(#triggered, me, #slotQuit)
  action.setShortcut("Ctrl+Q")
  menu.addAction(action)
end

--------------------------------------
--
--------------------------------------
on setupHelpActions(me)

  menuHelp = me.pMenuBar.addMenu("Help")

  action = $.Qt.Action.new(menuHelp)
  action.setText("About &"&me.pAppName)
  action.connect(#triggered, me, #slotAbout)
  --action.setShortcut("F1")
  menuHelp.addAction(action)

  action = $.Qt.Action.new(menuHelp)
  action.setText("About &Qt")
  action.connect(#triggered, me, #slotAboutQt)
  menuHelp.addAction(action)
end

--------------------------------------
-- @callback both action and event callback
--------------------------------------
on slotQuit (me)

  if (not me.maybeSave()) then return

  me.clearDateTimeWidget()

  --save window settings
  settings = $.Qt.Settings.new("Fluxus", me.pAppName, $.Qt.QSettings.IniFormat)
  settings.setValue("mainwindow/rect", string(me.pMainWindow.geometry()))
  settings.setValue("mainwindow/state", string(me.pMainWindow.windowState()))

  _movie.halt()
end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on slotAbout (me)
  msg = "MetadataEditor written in Lingo<br><br>"&\
  "This application is based on a my framework/runtime engine called 'LSW', "&\
  "it was written in a code editor (UltraEdit) without ever touching the Director IDE.<br><br>"&\
  "The idea of LSW is to write batch scripts (e.g.for daily automation tasks etc.) or "&\
  "desktop applications in Lingo, in a way similar to e.g. Node.js or Python.<br><br>"&\
  "More infos on LSW <a href='http://valentin.dasdeck.com/lingo/lsw/'>here</a>."

  $.Qt.MessageBox.about(me.pMainWindow, "About "&me.pAppName, msg)
end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on slotAboutQt (me)
  $.Qt.Application.aboutQt()
end

----------------------------------------
-- @callback
----------------------------------------
on cellPressed (me, widgetID, row, col)
  --put "cellPressed", widgetID, row, col
  me.clearDateTimeWidget()
end

----------------------------------------
-- @callback
----------------------------------------
on cellChanged (me, widgetID, row, col)
  --put "cellChanged", widgetID, row, col
  me.setModified(TRUE)
end

----------------------------------------
-- @callback
----------------------------------------
on cellDoubleClicked (me, widgetID, row, col)
  --put "cellDoubleClicked", widgetID, row, col

  n = me.pTabWidget.currentIndex() + 1

  tFileType = $.file.getFileType(me.pFileName)
  parserName = me.pFileTypeMap[tFileType]
  tTabs = me.pTabs[parserName]
  nsName = tTabs[n]
  tType = me.pNamespaces[nsName][#fields][row+1]

  if tType="date" then
    tableWidget = me.pTableWidgets[n]

    me.clearDateTimeWidget()

    me.pDateTimeWidget = $.Qt.DateTimeEdit.new()
    me.pDateTimeWidget.connect(#dateTimeChanged, me, #dateTimeChanged)

    -- get date of current cell
    tItem = tableWidget.item_(row, col)
    str = tItem.text() -- "2000-11-23T15:43:12"
    if str<>"" then
      res = $.date.parseISODateString(str)
      me.pDateTimeWidget.setDateTime(res.year, res.month, res.day, res.hour, res.minute, res.second)
    end if

    me.pDateTimeEditContainer = [tableWidget, row, col]
    tableWidget.setCellWidget(row, col, me.pDateTimeWidget)
  else
    me.clearDateTimeWidget()
  end if

end

----------------------------------------
-- @callback
----------------------------------------
on customContextMenuRequested (me, widgetID, p)
  me.pContextMenu.exec()
end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on slotFileOpen (me)

  if (not me.maybeSave()) then return

  -- create filter
  filter = ""
  repeat with ft in me.pFileTypes
    put "*."&ft&SPACE after filter
  end repeat
  delete the last char of filter
  filter = "Supported Media-Files ("&filter&")"

  fn = $.Qt.FileDialog.getOpenFileName (me.pMainWindow, "Open File...", $.PATH&"test_files", filter)

  if fn<>"" then
    fn = $.string.replace("/", $.PD, fn)
    me.load(fn)
  end if
end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on slotFileSave (me)
  me.save()
end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on slotFileImportMetadata (me)
  alert("TODO!")
end

--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on slotFileExportMetadata (me)
  alert("TODO!")
end


--------------------------------------
-- @callback object callback (called by action)
--------------------------------------
on slotFileLoadCSV (me)
  alert("TODO!")
end

--------------------------------------
--
--------------------------------------
on load (me, fn)

  tFileType = $.file.getFileType(fn)
  if pFileTypes.getPos(tFileType)=0 then return false

  parserName = me.pFileTypeMap[tFileType]
  parser = me.pParsers[parserName]
  info = parser.load(fn)

  if not listP(info) then
    put "Error: Parsing media file failed."
    return
  end if

  -- remove existing tabs
  cnt = me.pTabWidget.count_()
  repeat with i = cnt-1 down to 0
    me.pTabWidget.removeTab(i)
  end repeat

  -- create metadata tabs
  me.pTableWidgets = []

  repeat with nsName in me.pTabs[parserName]
    ns = me.pNamespaces[nsName]
    tableWidget = me.createTable(ns)

    me.pTabWidget.addTab(tableWidget, ns.name)
    me.pTableWidgets.add(tableWidget)

    --tableWidget.setColumnWidth(0, 320)
    --tableWidget.setColumnWidth(1, 320)

    if not voidP(info[nsName]) then

      -- load data into tabs
      fields = ns.fields
      cntFields = fields.count
      repeat with i = 1 to cntFields

        tProp = fields.getPropAt(i)

        tValue = info[nsName][tProp]
        if voidP(tValue) then next repeat

        -- get datatype
        dataType = fields[i] -- string
        if voidP(dataType) then dataType="string"

        if listP(tValue) then
          tmp = ""
          repeat with str in tValue
            put str&";" after tmp
          end repeat
          delete the last char of tmp
          tValue = tmp
        end if

        -- tmp
        tValue = string(tValue)

        row = i-1

        tItem = tableWidget.item_(row, 0)
        tItem.setText(tProp)

        tItem = tableWidget.item_(row, 1)
        tItem.setText(tValue)

      end repeat

    end if

    tableWidget.connect(#cellPressed, me, #cellPressed)
    tableWidget.connect(#cellChanged, me, #cellChanged)

  end repeat

  -- force view update
  me.pTabWidget.setVisible(0)
  me.pTabWidget.setVisible(1)

  me.setCurrentFileName(fn)

  return TRUE
end

--------------------------------------
--
--------------------------------------
on save (me)

  -- exit table edit mode to also save cell with edit focus
  n = me.pTabWidget.currentIndex() + 1
  tableWidget = me.pTableWidgets[n]
  tableWidget.setCurrentCell(-1, -1)

  -- get current parser
  tFileType = $.file.getFileType(me.pFileName)
  parserName = me.pFileTypeMap[tFileType]
  parser = me.pParsers[parserName]

  -- create metadata proplist
  info = [:]
  info["namespaces"] = [:]

  cnt = me.pTabs[parserName].count
  repeat with i = 1 to cnt
    nsName = me.pTabs[parserName][i]
    tableWidget = me.pTableWidgets[i]

    cntRows = tableWidget.rowCount()
    repeat with row = 1 to cntRows
      prop = tableWidget.item_(row-1, 0).text()
      if string(prop)="" then exit repeat

      val = tableWidget.item_(row-1, 1).text()
      if string(val)<>"" then

        if voidP(info[nsName]) then
          info["namespaces"][nsName] = me.pNamespaces[nsName].url
          info[nsName] = [:]
        end if

        -- get datatype
        fields = me.pNamespaces[nsName].fields

        dataType = fields[prop] -- string
        if voidP(dataType) then dataType="string"

        if dataType="list" then
          val = $.string.explodeBySingleChar(";", val)
        end if

        info[nsName][prop] = val
      end if

    end repeat

  end repeat

  -- save changes
  ok = parser.save(me.pFileName, info)
  if ok then
    put "Changes saved successfully."
    me.setModified(FALSE)
  else
    put "Error: Changes couldn't be saved."
  end if

  return ok
end

--------------------------------------
--
--------------------------------------
on setCurrentFileName(me, fn)
  me.pFileName = fn

  -- extract filename
  tmp = $.string.explodeBySingleChar($.PD, fn)
  shownName = tmp[tmp.count]
  me.pMainWindow.setWindowTitle(me.pAppName&" - ["&shownName&"[*]]")

  me.setModified(FALSE)

  me.pActionImportMetadata.setEnabled(TRUE)
  me.pActionExportMetadata.setEnabled(TRUE)
end

--------------------------------------
--
--------------------------------------
on setModified (me, flag)
  me.pIsModified = flag
  me.pActionSave.setEnabled(flag)
  me.pMainWindow.setWindowModified(flag)
end

--------------------------------------
--
--------------------------------------
on maybeSave (me)
  if (me.pIsModified) then

    msgBox = $.Qt.MessageBox.new()
    msgBox.setIcon(4)
    msgBox.setWindowTitle(me.pAppName)
    msgBox.setText("The current file's metadata has been modified."&$.LF&"Do you want to save your changes?")
    msgBox.setWindowIcon($.@("assets/appicon.png"))

    flags = $.Qt.QMessageBox.Save
    flags = bitOr(flags, $.Qt.QMessageBox.Discard)
    flags = bitOr(flags, $.Qt.QMessageBox.Cancel)

    msgBox.setStandardButtons(flags)

    msgBox.setDefaultButton($.Qt.QMessageBox.Save)

    ret = msgBox.exec()

    if (ret = $.Qt.QMessageBox.Save) then
      return me.save()
    else if (ret = $.Qt.QMessageBox.Cancel) then
      return FALSE
    end if

  end if
  return TRUE
end

----------------------------------------
-- @callback dropHandler
----------------------------------------
on dropHandler(me, resList, point)

  if (not me.maybeSave()) then return

  if $.os="win" then
    prop = "application/x-qt-windows-mime;value="&QUOTE&"FileNameW"&QUOTE
    data = resList[prop]
    if not voidP(data) then
      data.position = 1
      fn = data.readRawString(data.length, "utf-16")
      me.load(fn)
    end if

  else
    prop = "text/uri-list"
    data = resList[prop]
    if not voidP(data) then
      data.position = 1
      fn = data.readRawString(data.length)
      fn = fn.line[1]
      delete char 1 to 7 of fn -- file://
      fn = qx_posix2hfs(fn)
      me.load(fn)
    end if

  end if
end

----------------------------------------
--
----------------------------------------
on clearDateTimeWidget (me)
  if not voidP(me.pDateTimeWidget) then
    tableWidget = me.pDateTimeEditContainer[1]
    row = me.pDateTimeEditContainer[2]
    col = me.pDateTimeEditContainer[3]
    tableWidget.removeCellWidget(row, col)
    me.pDateTimeWidget.destroy()
    me.pDateTimeWidget = VOID
  end if
end

----------------------------------------
--
----------------------------------------
on dateTimeChanged (me, widgetID, dateTime)
  --put "dateTimeChanged", dateTime -- "20001123 15:43:12"

  tmp = $.string.explodeBySingleChar(" ", dateTime)

  date = tmp[1]
  time = tmp[2]

  year = date.char[1..4]
  month = date.char[5..6]
  day = date.char[7..8]

  tmp = $.string.explodeBySingleChar(":", time)
  h = tmp[1]
  m = tmp[2]
  s = tmp[3]

  -- me.pDateTimeEditContainer = [tableWidget, row, col]
  tableWidget = me.pDateTimeEditContainer[1]
  row = me.pDateTimeEditContainer[2]
  col = me.pDateTimeEditContainer[3]

  tItem = tableWidget.item_(row, col)
  strOld = tItem.text()
  --strNew = year&"-"&month&"-"&day&"T"&h&":"&m&":"&s
  strNew = $.date.createISODateString(year,month,day,h,m,s)
  if strNew<>strOld then
    tItem.setText(strNew)
    me.setModified(TRUE)
  end if
end

  on   foo
    --
  end